/*******************************************************************************
 *	ChaseDa.c	CHASE_DA Linux Device Driver
 *			T.Barnaby,	BEAM Ltd,	2009-04-21
 *******************************************************************************
 * Copyright (c) 2009 BEAM Ltd. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
 
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include "chase_da.h"

#define	VERSION			"1.0.2"
#define NAME			"chase_da"
#define NAME_PREFIX		"chase_da: "
#define MAX_NUM_CARDS		8

// Delay value
#define CHASE_DA_DELAY		5

// The board registers
#define	CHASE_DA_REG_0		0
#define	CHASE_DA_REG_1		1

typedef struct {
	unsigned long		physAddress;
	unsigned		len;
	void __iomem*		address;
} MemMap;

typedef struct DeviceData_struct {
	int			boardType;
	int			opened;
	dev_t			dev;
	struct pci_dev*		pdev;
	struct cdev		cdev;
	struct device*		device;
	MemMap			regs;
	MemMap			mem;
	uint16_t		regSave[2];
} DeviceData;

struct class*			class;
static int			cardsNum;
static DeviceData		cards[MAX_NUM_CARDS];
static dev_t			firstDev;
static spinlock_t		nextDevLock;

static struct pci_device_id chase_da_ids[] = {
	{ PCI_VENDOR_ID_PLX, 0x9031, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA11000 },
	{ PCI_VENDOR_ID_PLX, 0x9033, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA11000 },
//	{ PCI_VENDOR_ID_PLX, 0x9053, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA4310 },
	{ PCI_VENDOR_ID_PLX, 0x9052, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA8150 },

	{ 0x1197, 0x0301, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA11000 },
	{ 0x1197, 0x0311, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA11000 },
//	{ 0x1197, 0x0302, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA4310 },
//	{ 0x1197, 0x0312, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA4310 },
	{ 0x1197, 0x0303, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA8150 },
	{ 0x1197, 0x0313, PCI_ANY_ID, PCI_ANY_ID, 0, 0, CHASE_DA8150 },

	{ 0, }
};
MODULE_DEVICE_TABLE(pci, chase_da_ids);

static void reg_write(DeviceData* devData, uint32_t address, uint16_t data){
	devData->regSave[address] = data;
	((volatile uint16_t*)devData->regs.address)[address] = data;
}

#ifdef ZAP
static uint16_t reg_read(DeviceData* devData, uint32_t address){
	return ((volatile uint16_t*)devData->regs.address)[address];
}
#endif

static void chase_da_memEnable(DeviceData* devData, uint32_t on){
	if(on)
		reg_write(devData, 0, devData->regSave[0] | 0x4000);
	else
		reg_write(devData, 0, devData->regSave[0] & ~0x4000);
}

static void chase_da_setTriggerMode(DeviceData* devData, uint8_t mode, uint8_t extPol){
	if(mode == 0){
		reg_write(devData, 1, devData->regSave[1] | 0x1);       // Shuts down MAD Counter
		reg_write(devData, 0, devData->regSave[0] & 0x7FFF);  // Shuts down TRIG_EN if on

		reg_write(devData, 1, devData->regSave[1] | 0x0080);  // SYNC_DAC_TTL
		reg_write(devData, 1, devData->regSave[1] & 0xEFFF);  // Disables VCO
	// if (ClockRate < 60000000) SetDDS(0);  // Disables DDS if selected
	}

	if((mode == 1) || (mode == 2)){
		// SYNCHRONIZE UPCONVERTERS
		reg_write(devData, 1, devData->regSave[1] | 0x1);       // Shuts down MAD Counter

		reg_write(devData, 1, devData->regSave[1] & 0xEFFF);  // Disables VCO
		// if (ClockRate < 60000000) SetDDS(0);  // Disables DDS if selected

		reg_write(devData, 0, devData->regSave[0] | 0x2000);  // SYNC_RAM_TTL
		reg_write(devData, 0, devData->regSave[0] & 0xDFFF);  // Toggles High to Reset

		reg_write(devData, 1, devData->regSave[1] | 0x0080);  // SYNC_DAC_TTL
		reg_write(devData, 1, devData->regSave[1] & 0xFF7F);  // Toggles HIGH to Reset


		if(mode == 2){
			reg_write(devData, 0, devData->regSave[0] | 0x8000);  // Activates TRIG_EN
			if(extPol == 1){
				reg_write(devData, 1, devData->regSave[1] | 0x0100);  // Sets TRIG_POL for Falling Edge(inverts TRIG_IN)
			}
			else{
				reg_write(devData, 1, devData->regSave[1] & 0xFEFF);  // Sets TRIG_POL for Rising Edge
			}
		}
		else{
			reg_write(devData, 0, devData->regSave[0] & 0x7FFF);  // Shuts down TRIG_EN
		}

		reg_write(devData, 1, devData->regSave[1] | 0x1000);  // Enables VCO
		// if (ClockRate < 60000000) SetDDS(ClockRate);  // Enables DDS if selected

		reg_write(devData, 1, devData->regSave[1] & 0xFFFE);    // Enables MAD Counter.
	}
}

static void chase_da_setRamOffset(DeviceData* devData, uint16_t offset)
{
	offset = offset << 9;
	reg_write(devData, 1, (devData->regSave[1] & 0xF1FF));		// Clear RAM_Offset bits
	reg_write(devData, 1, (devData->regSave[1] | offset));		// Apply RAM_Offset
}

static void chase_da_setAttenuation(DeviceData* devData, uint32_t chanNum, uint32_t value)
{
	int	x;
	
	value = value & 0x3F;	// Need only 6-bits

	// Set ALL LE Low to Allow Serial Data + Set SER_DATA/CLOCK Low
	if(devData->boardType == CHASE_DA8150)
		reg_write(devData, 0, (devData->regSave[0] & 0xFFC0));
	else
		reg_write(devData, 0, (devData->regSave[0] & 0xFFE0));

	// Clock in Data Starting with MSB First
	for(x = 0; x < 6; x++){
		if((value & 0x20) > 0)
			reg_write(devData,  0,(devData->regSave[0] | 0x1));  // Set SER_DATA to 1
		else
			reg_write(devData,  0,(devData->regSave[0] & 0xFFFE)); // Set SER_DATA to 0

		reg_write(devData, 0, (devData->regSave[0] | 0x2)); reg_write(devData, 0, (devData->regSave[0] & 0xFFFD));  // Clock _|^|_
		value = value << 1;
	}

	// Create Latch Pulse _|^|_
	if(chanNum == 0x01){ reg_write(devData, 0, (devData->regSave[0] | 0x4 )); reg_write(devData, 0, (devData->regSave[0] & 0xFFFB)); }
	if(chanNum == 0x02){ reg_write(devData, 0, (devData->regSave[0] | 0x8 )); reg_write(devData, 0, (devData->regSave[0] & 0xFFF7)); }
	if(devData->boardType == CHASE_DA8150){
		if(chanNum == 0x04){ reg_write(devData, 0, (devData->regSave[0] | 0x10)); reg_write(devData, 0, (devData->regSave[0] & 0xFFEF)); }
		if(chanNum == 0x08){ reg_write(devData, 0, (devData->regSave[0] | 0x20)); reg_write(devData, 0, (devData->regSave[0] & 0xFFDF)); }
	}
}

static int chase_da_setClock(DeviceData* devData, uint32_t frequency)
{
	int	ret = 0;
	
	if(devData->boardType == CHASE_DA8150){
		if(frequency > 300000000){
			reg_write(devData, 0, (devData->regSave[0] & 0xF03F) | 0x0800 );	// Set Clock To External For Debug + Shut down TTL Clk
		}
		if(frequency == 150000000){
			reg_write(devData, 0, (devData->regSave[0] & 0xF03F) | 0x0400 );	// Set Clock To Internal 300 MHz + DIV/1
		}
		if(frequency == 75000000){
			reg_write(devData, 0, (devData->regSave[0] & 0xF03F) | 0x0040 );	// Set Clock To Internal 150 MHz + DIV/2
		}
		if(frequency == 37500000){
			reg_write(devData, 0, (devData->regSave[0] & 0xF03F) | 0x0080 );	// Set Clock To Internal 75 MHz + DIV/4
		}


	//	if(frequency >= 1500e6){
	//		reg_write(devData, 0, (devData->regSave[0] & 0xF03F) | 0x800 );  // Set Clock Mux to External Clock + DIV/1
	//		return;
	//	}
	}
	else {
		ret = -EINVAL;
	}
	
	return ret;
}

static void chase_da_setSoftTrigger(DeviceData* devData){
	if((devData->regSave[1] & 0x0100) > 0 ){
		reg_write(devData, 1, devData->regSave[1] & 0xFEFF);
		reg_write(devData, 1, devData->regSave[1] | 0x0100);
	}
	else{
		reg_write(devData, 1, devData->regSave[1] | 0x0100);
		reg_write(devData, 1, devData->regSave[1] & 0xFEFF);
	}
}

static void chase_da_load_pll(DeviceData* devData, uint32_t value){
	int	i;
	
	// LMX2316 PROGRAMMING
	// LSB                                                       SHIFTED IN FIRST ==>  MSB
	// C1  C2  F1  F2  F3  F4  F5  F6  F7  F8  F9  F10 F11 F12 F13 F14 F15 F16 F17 F18 F19
	
	reg_write(devData, 0, (devData->regSave[0] & 0xFFFC));	// Set SER_DATA & SER_CLOCK to zero
	udelay(CHASE_DA_DELAY);
	reg_write(devData, 1, (devData->regSave[1] & 0xFFEF));	// Set PLL_LD_EN to Zero - Allows clocking data into PLL
	udelay(CHASE_DA_DELAY);

	for(i = 0; i < 21; i++){
		// Set data bit
		if((value & 0x100000) > 0)
			reg_write(devData, 0, (devData->regSave[0] | 0x0001));
		else
			reg_write(devData, 0, (devData->regSave[0] & 0xFFFE));

		udelay(CHASE_DA_DELAY);
		
		// Clock data
		reg_write(devData, 0, (devData->regSave[0] | 0x0002));
		reg_write(devData, 0, (devData->regSave[0] & 0xFFFD));
		value = value << 1;
	}

	// Load register values into device
	udelay(CHASE_DA_DELAY);
	reg_write(devData, 1, (devData->regSave[1] | 0x0010));
	udelay(CHASE_DA_DELAY);
}


static int chase_da_open(struct inode* inode, struct file* filp){
	int		ret = 0;
	DeviceData*	devData;

	devData = container_of(inode->i_cdev, DeviceData, cdev);
	spin_lock(&nextDevLock);
	if(devData->opened){
		spin_unlock(&nextDevLock);
		return -EBUSY;
	}
	devData->opened = 1;
	spin_unlock(&nextDevLock);
	
	devData->regSave[0] = 0x4900;	// Initialize ghost registers
	devData->regSave[1] = 0x0201;	// Sync Reset, Set RAM_Offset=1
	reg_write(devData, 0, devData->regSave[0]);
	reg_write(devData, 1, devData->regSave[1]);

	if(devData->boardType == CHASE_DA11000){
		uint32_t	pllFunction	= 0x000092;   // 0x000092
		uint32_t	pllRCounter	= 0x100028;   // 0x100050 => 20MHz, 0x100028 => 10MHz
		uint32_t	pllNCounter	= 0x000FA1;   // 0x100FA1

		// LMX2316 PROGRAMMING
		// LOAD FUNCTION LATCH (0x10), THEN R COUNTER (0x00), THEN N COUNTER (0x01)
		chase_da_load_pll(devData, pllFunction);
		chase_da_load_pll(devData, pllRCounter);
		chase_da_load_pll(devData, pllNCounter);

		chase_da_memEnable(devData, 0);
		chase_da_setTriggerMode(devData, 0, 0);
		chase_da_setRamOffset(devData, 3);

		reg_write(devData, 0, (devData->regSave[0] & 0xFFCF));	// Clear CLK_Offset bits
		reg_write(devData, 0, (devData->regSave[0] | 0x0010));	// CLK_Offset
	}
	else {
		chase_da_memEnable(devData, 0);
		chase_da_setTriggerMode(devData, 0, 0);
		chase_da_setRamOffset(devData, 2);
	}

	return ret;
}

static int chase_da_release(struct inode* inode, struct file* filp){
	DeviceData*	devData;

	devData = container_of(inode->i_cdev, DeviceData, cdev);
	devData->opened = 0;
	
	return 0;
}

static int chase_da_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg){
	int			ret = 0;
	ChaseDaInfo		info;
	ChaseDaAttenuation	atten;
	DeviceData*		devData;

	devData = container_of(inode->i_cdev, DeviceData, cdev);

	switch(cmd){
	case CHASE_DA_CMD_GETINFO:
		info.boardType = devData->boardType;
		info.regsPhysAddress = devData->regs.physAddress;
		info.regsLen = devData->regs.len;
		info.memPhysAddress = devData->mem.physAddress;
		info.memLen = devData->mem.len;
		unlikely(copy_to_user((void __user*)arg, &info, sizeof(info)));
		break;
	case CHASE_DA_CMD_MEM_ENABLE:
		chase_da_memEnable(devData, arg);
		break;
	case CHASE_DA_CMD_SET_CLOCK:
		ret = chase_da_setClock(devData, arg);
		break;
	case CHASE_DA_CMD_SET_TRIGGER:
		chase_da_setTriggerMode(devData, (arg >> 8) & 0xFF, (arg) & 0xFF);
		break;
	case CHASE_DA_CMD_SOFT_TRIGGER:
		chase_da_setSoftTrigger(devData);
		break;
	case CHASE_DA_CMD_SET_ATTENUATION:
		unlikely(copy_from_user(&atten, (void __user*)arg, sizeof(atten)));
		chase_da_setAttenuation(devData, atten.chan, atten.value);
		break;
	default:
		ret = -EINVAL;
	}
	return ret;
}

static int chase_da_mmap(struct file* file, struct vm_area_struct* vma){
	DeviceData*	data;
	unsigned long	off;
	unsigned long	len;
	int		ret;

	data = container_of(file->f_dentry->d_inode->i_cdev, DeviceData, cdev);

	off = vma->vm_pgoff << PAGE_SHIFT;
	
	if(off == data->regs.physAddress){
		len = data->regs.len;
		len = 4096;
	}
	else if(off == data->mem.physAddress){
		len = data->mem.len;
	}
	else {
		return -EINVAL;
	}
	
	if((vma->vm_end - vma->vm_start) > len)
		return -EINVAL;

	/* Ensure mapping is suitable for IO */
	vma->vm_flags |= VM_IO | VM_RESERVED;
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

#if LINUX_VERSION_CODE <=  KERNEL_VERSION(2,6,12) 	 
	ret = io_remap_page_range(vma, vma->vm_start, off, len, vma->vm_page_prot);
#else 	 
	ret = remap_pfn_range(vma, vma->vm_start, (unsigned long)(off)>>PAGE_SHIFT, len, vma->vm_page_prot);
#endif 	 
 
	if (ret)
		return -EAGAIN;

	return 0;
}

#ifdef ZAP
static irqreturn_t chase_da_intr_handler(int irq, void *arg, struct pt_regs *regs){
	return 0;
}
#endif

static struct file_operations fops = {
	.owner = THIS_MODULE,
	open:		chase_da_open,
	release:	chase_da_release,
	ioctl:		chase_da_ioctl,
	mmap:		chase_da_mmap,
};

static int map_region(struct pci_dev* pdev, int region, MemMap* map){
	/* Map in memory region */
	map->physAddress = pci_resource_start(pdev, region);
	map->len = pci_resource_len(pdev, region);
	if(!map->physAddress){
		printk(KERN_ERR NAME_PREFIX "PCI region %d not available\n", region);
		return -ENODEV;
	}

	map->address = ioremap_nocache(map->physAddress, map->len);
	if(!map->address){
		printk(KERN_ERR NAME_PREFIX "Cannot map PCI Region: %d Address: 0x%lx: Len: %u\n", region, map->physAddress, map->len);
		release_mem_region(map->physAddress, map->len);
		return -ENODEV;
	}
	return 0;
}

static int unmap_region(struct pci_dev* pdev, int region, MemMap* map){
	/* Unmap memory region */
	iounmap(map->address);

	return 0;
}

static int chase_da_probe(struct pci_dev* pdev, const struct pci_device_id* id)
{
	int			ret;
	DeviceData*		data;
	dev_t			dev = firstDev + cardsNum;

	if(pci_enable_device(pdev))
		return -EIO;

	if((ret = pci_request_regions(pdev, NAME)) < 0)
		return -ENOMEM;

	/* Allocate a new card */
	spin_lock(&nextDevLock);
	if((cardsNum + 1) >= MAX_NUM_CARDS){
		spin_unlock(&nextDevLock);
		return -ENODEV;
	}
	data = &cards[cardsNum];
	cardsNum++;
	spin_unlock(&nextDevLock);
	
	/* Setup card information */
	pci_set_drvdata(pdev, data);
	data->dev = dev;
	data->pdev = pdev;
	data->boardType = id->driver_data;

	/* Map in memory regions */
	if((ret = map_region(pdev, 3, &data->regs)))
		return ret;

	if((ret = map_region(pdev, 2, &data->mem)))
		return ret;

#ifdef ZAP
	/* Register interrupt handler */
	if((ret = request_irq(pdev->irq, chase_da_intr_handler, SA_SHIRQ, NAME, data))){
		printk(KERN_ERR NAME_PREFIX "Cannot get interrupt: %d\n", pdev->irq);
		return -ENODEV;
	}
#endif

	/* Setup device */
	data->cdev.owner = THIS_MODULE;
	kobject_set_name(&data->cdev.kobj, "chase_da-%d", MINOR(dev));
	cdev_init(&data->cdev, &fops);
	if((ret = cdev_add(&data->cdev, dev, 1))){
		printk(KERN_ERR NAME_PREFIX "Cannot add cdev\n");
		return -ENODEV;
	}

	/* Setup /dev entry */	
	data->device = device_create(class, NULL, dev, "chase_da-%d", MINOR(dev)); 
	if(data->device)
		dev_set_drvdata(data->device, data);

	printk(KERN_INFO NAME_PREFIX "Driver loaded: Version: %s\n", VERSION);

	return 0;
}

static void chase_da_remove(struct pci_dev *pdev)
{
	DeviceData*	data;

	/* clean up any allocated resources and stuff here.
	 * like call release_region();
	 */
	data = pci_get_drvdata(pdev);

	device_destroy(class, data->dev);
	cdev_del(&data->cdev);

	unmap_region(pdev, 1, &data->mem);
	unmap_region(pdev, 0, &data->regs);
	pci_release_regions(pdev);

	printk(KERN_INFO NAME_PREFIX "Driver unloaded\n");
}

static struct pci_driver pci_driver = {
	.name		= "chase_da",
	.id_table	= chase_da_ids,
	.probe		= chase_da_probe,
	.remove		= chase_da_remove,
};

static int __devinit chase_da_init(void)
{
	int	ret;

	nextDevLock = SPIN_LOCK_UNLOCKED;

	// Allocate Device
	if((ret = alloc_chrdev_region(&firstDev, 0, 1, NAME))){
		printk(KERN_ERR NAME_PREFIX "Failed to allocate device number block\n");
		return ret;
	}

	// Create class for device	
	if(IS_ERR(class = class_create(THIS_MODULE, NAME)))
		return PTR_ERR(class);

	return pci_register_driver(&pci_driver);
}

static void __exit chase_da_exit(void)
{
	/* Unregister device driver */
	pci_unregister_driver(&pci_driver);

	class_destroy(class);

	/* Release device number block */
	unregister_chrdev_region(firstDev, 1);
}

module_init(chase_da_init);
module_exit(chase_da_exit);

MODULE_AUTHOR("BEAM Ltd");
MODULE_DESCRIPTION("Chase2000 CHASE_DA device driver");
MODULE_LICENSE("GPL");
